import Combine
import Foundation
import UIKit
import Storage
import SwiftUI
import Yosemite

/// Facilitates connecting to a card reader
///
final class BuiltInCardReaderConnectionController {
    private enum ControllerState {
        /// Initial state of the controller
        ///
        case idle

        /// Initializing (fetching payment gateway accounts)
        ///
        case initializing

        /// Preparing for search
        ///
        case preparingForSearch

        /// Begin search for card readers
        ///
        case beginSearch

        /// Searching for a card reader
        ///
        case searching

        /// Attempting to connect to a card reader. The completion passed to `searchAndConnect`
        /// will be called with a `success` `Bool` `True` result if successful, after which the view controller
        /// passed to `searchAndConnect` will be dereferenced and the state set to `idle`
        ///
        case connectToReader

        /// A failure occurred while connecting. The search may continue or be canceled. At this time we
        /// do not present the detailed error from the service.
        ///
        case connectingFailed(Error)

        /// A mandatory update is being installed
        ///
        case updating(progress: Float)

        /// User chose to retry the connection to the card reader. Starts the search again, by dismissing modals and initializing from scratch
        ///
        case retry

        /// User cancelled search/connecting to a card reader. The completion passed to `searchAndConnect`
        /// will be called with a `success` `Bool` `False` result. The view controller passed to `searchAndConnect` will be
        /// dereferenced and the state set to `idle`
        ///
        case cancel

        /// A failure occurred. The completion passed to `searchAndConnect`
        /// will be called with a `failure` result. The view controller passed to `searchAndConnect` will be
        /// dereferenced and the state set to `idle`
        ///
        case discoveryFailed(Error)
    }

    private let storageManager: StorageManagerType
    private let stores: StoresManager

    private var state: ControllerState {
        didSet {
            didSetState()
        }
    }

    private let siteID: Int64
    private let alertsPresenter: CardPresentPaymentAlertsPresenting
    private let configuration: CardPresentPaymentsConfiguration

    private let alertsProvider: CardReaderConnectionAlertsProviding

    /// The reader we want the user to consider connecting to
    ///
    private var candidateReader: CardReader?

    /// Tracks analytics for card reader connection events
    ///
    private let analyticsTracker: CardReaderConnectionAnalyticsTracker

    private var softwareUpdateCancelable: FallibleCancelable? = nil

    private var subscriptions = Set<AnyCancellable>()

    private var onCompletion: ((Result<CardReaderConnectionResult, Error>) -> Void)?

    private(set) lazy var dataSource: CardReaderSettingsDataSource = {
        return CardReaderSettingsDataSource(siteID: siteID, storageManager: storageManager)
    }()

    /// Gateway ID to include in tracks events
    private var gatewayID: String? {
        didSet {
            didSetGatewayID()
            analyticsTracker.setGatewayID(gatewayID: gatewayID)
        }
    }

    init(
        forSiteID: Int64,
        storageManager: StorageManagerType = ServiceLocator.storageManager,
        stores: StoresManager = ServiceLocator.stores,
        alertsPresenter: CardPresentPaymentAlertsPresenting,
        alertsProvider: CardReaderConnectionAlertsProviding,
        configuration: CardPresentPaymentsConfiguration,
        analyticsTracker: CardReaderConnectionAnalyticsTracker
    ) {
        siteID = forSiteID
        self.storageManager = storageManager
        self.stores = stores
        state = .idle
        self.alertsPresenter = alertsPresenter
        self.alertsProvider = alertsProvider
        self.configuration = configuration
        self.analyticsTracker = analyticsTracker

        configureResultsControllers()
        loadPaymentGatewayAccounts()
    }

    deinit {
        subscriptions.removeAll()
    }

    func searchAndConnect(onCompletion: @escaping (Result<CardReaderConnectionResult, Error>) -> Void) {
        self.onCompletion = onCompletion
        self.state = .initializing
    }
}

private extension BuiltInCardReaderConnectionController {
    func configureResultsControllers() {
        dataSource.configureResultsControllers(onReload: { [weak self] in
            guard let self = self else { return }
            self.gatewayID = self.dataSource.cardPresentPaymentGatewayID()
        })
        // Sets gateway ID from initial fetch.
        gatewayID = dataSource.cardPresentPaymentGatewayID()
    }

    func loadPaymentGatewayAccounts() {
        let action = CardPresentPaymentAction.loadAccounts(siteID: siteID) {_ in}
        stores.dispatch(action)
    }

    func didSetState() {
        switch state {
        case .idle:
            onIdle()
        case .initializing:
            onInitialization()
        case .preparingForSearch:
            onPreparingForSearch()
        case .beginSearch:
            onBeginSearch()
        case .searching:
            onSearching()
        case .retry:
            onRetry()
        case .cancel:
            onCancel()
        case .connectToReader:
            onConnectToReader()
        case .connectingFailed(let error):
            onConnectingFailed(error: error)
        case .discoveryFailed(let error):
            onDiscoveryFailed(error: error)
        case .updating(progress: let progress):
            onUpdateProgress(progress: progress)
        }
    }

    /// Once the gatewayID arrives (during initialization) it is OK to proceed with search preparations
    ///
    func didSetGatewayID() {
        if case .initializing = state {
            state = .preparingForSearch
        }
    }

    /// Initial state of the controller
    ///
    func onIdle() {
    }

    /// Searching for a reader is about to begin. Wait, if needed, for the gateway ID to be provided from the FRC
    ///
    func onInitialization() {
        if gatewayID != nil {
            state = .preparingForSearch
        }
    }

    /// In preparation for search, sets everything to defaults
    /// Does NOT open any modal
    /// Transitions state to `.beginSearch`
    ///
    func onPreparingForSearch() {
        /// Always start fresh
        ///
        candidateReader = nil

        if case .preparingForSearch = state {
            state = .beginSearch
        }
    }

    /// Begins the search for a card reader
    /// Does NOT open any modal
    /// Transitions state to `.searching`
    /// Later, when a reader is found, state transitions to `.connectToReader`
    ///
    func onBeginSearch() {
        self.state = .searching

        let action = CardPresentPaymentAction.startCardReaderDiscovery(
            siteID: siteID,
            discoveryMethod: .localMobile,
            onReaderDiscovered: { [weak self] cardReaders in
                guard let self = self else {
                    return
                }

                /// Note: This completion will be called repeatedly as the list of readers
                /// discovered changes, so some care around state must be taken here.
                ///

                /// If we have a found reader, advance to `connectToReader`
                ///
                if cardReaders.isNotEmpty {
                    self.candidateReader = cardReaders.first
                    self.state = .connectToReader
                    return
                }
            },
            onError: { [weak self] error in
                guard let self = self else { return }

                ServiceLocator.analytics.track(
                    event: WooAnalyticsEvent.InPersonPayments.cardReaderDiscoveryFailed(forGatewayID: self.gatewayID,
                                                                                        error: error,
                                                                                        countryCode: self.configuration.countryCode)
                )
                self.state = .discoveryFailed(error)
            })

        stores.dispatch(action)
    }

    /// Opens the scanning for reader modal
    /// If the user cancels the modal will trigger a transition to `.endSearch`
    ///
    func onSearching() {
        /// Display the single view and ask the merchant if they'd
        /// like to connect to it
        ///
        if candidateReader != nil {
            self.state = .connectToReader
            return
        }

        /// If all else fails, display the "scanning" modal and
        /// stay in this state
        ///
        alertsPresenter.present(viewModel: alertsProvider.scanningForReader(cancel: {
            self.state = .cancel
        }))
    }

    /// A mandatory update is being installed
    ///
    func onUpdateProgress(progress: Float) {
        let cancel = softwareUpdateCancelable.map { cancelable in
            return { [weak self] in
                guard let self = self else { return }
                self.state = .cancel
                self.analyticsTracker.cardReaderSoftwareUpdateCancelTapped()
                cancelable.cancel { [weak self] result in
                    if case .failure(let error) = result {
                        DDLogError("💳 Error: canceling software update \(error)")
                    } else {
                        self?.analyticsTracker.cardReaderSoftwareUpdateCanceled()
                    }
                }
            }
        }

        alertsPresenter.present(
            viewModel: alertsProvider.updateProgress(requiredUpdate: true,
                                                     progress: progress,
                                                     cancel: cancel))
    }

    /// Retry a search for a card reader
    ///
    func onRetry() {
        alertsPresenter.dismiss()
        let action = CardPresentPaymentAction.cancelCardReaderDiscovery() { [weak self] _ in
            self?.state = .beginSearch
        }
        stores.dispatch(action)
    }

    /// End the search for a card reader
    ///
    func onCancel() {
        let action = CardPresentPaymentAction.cancelCardReaderDiscovery() { [weak self] _ in
            self?.returnSuccess(result: .canceled)
        }
        stores.dispatch(action)
    }

    /// Connect to the candidate card reader
    ///
    func onConnectToReader() {
        guard let candidateReader = candidateReader else {
            return
        }

        analyticsTracker.setCandidateReader(candidateReader)

        let softwareUpdateAction = CardPresentPaymentAction.observeCardReaderUpdateState { [weak self] softwareUpdateEvents in
            guard let self = self else { return }

            softwareUpdateEvents.sink { [weak self] event in
                guard let self = self else { return }

                switch event {
                case .started(cancelable: let cancelable):
                    self.softwareUpdateCancelable = cancelable
                    self.state = .updating(progress: 0)
                case .installing(progress: let progress):
                    if progress >= 0.995 {
                        self.softwareUpdateCancelable = nil
                    }
                    self.state = .updating(progress: progress)
                case .completed:
                    self.softwareUpdateCancelable = nil
                    self.state = .updating(progress: 1)
                default:
                    break
                }
            }
            .store(in: &self.subscriptions)
        }
        stores.dispatch(softwareUpdateAction)

        let action = CardPresentPaymentAction.connect(reader: candidateReader) { [weak self] result in
            guard let self = self else { return }

            self.analyticsTracker.setCandidateReader(nil)

            switch result {
            case .success(let reader):
                ServiceLocator.analytics.track(
                    event: WooAnalyticsEvent.InPersonPayments
                        .cardReaderConnectionSuccess(forGatewayID: self.gatewayID,
                                                     batteryLevel: reader.batteryLevel,
                                                     countryCode: self.configuration.countryCode,
                                                     cardReaderModel: reader.readerType.model)
                )
                // If we were installing a software update, introduce a small delay so the user can
                // actually see a success message showing the installation was complete
                if case .updating(progress: 1) = self.state {
                    DispatchQueue.main.asyncAfter(deadline: .now() + .seconds(1)) {
                        self.returnSuccess(result: .connected(reader))
                    }
                } else {
                    self.returnSuccess(result: .connected(reader))
                }
            case .failure(let error):
                ServiceLocator.analytics.track(
                    event: WooAnalyticsEvent.InPersonPayments.cardReaderConnectionFailed(forGatewayID: self.gatewayID,
                                                                                         error: error,
                                                                                         countryCode: self.configuration.countryCode,
                                                                                         cardReaderModel: candidateReader.readerType.model)
                )
                self.state = .connectingFailed(error)
            }
        }
        stores.dispatch(action)

        alertsPresenter.present(viewModel: alertsProvider.connectingToReader())
    }

    /// An error occurred while connecting
    ///
    private func onConnectingFailed(error: Error) {
        /// Clear our candidateReader to avoid connecting to a reader that isn't
        /// there while we wait for `onReaderDiscovered` to receive an update.
        /// See also https://github.com/stripe/stripe-terminal-ios/issues/104#issuecomment-916285167
        ///
        self.candidateReader = nil

        if case CardReaderServiceError.softwareUpdate(underlyingError: let underlyingError, batteryLevel: _) = error,
           underlyingError.isSoftwareUpdateError {
            return onUpdateFailed(error: error)
        }
        showConnectionFailed(error: error)
    }

    private func onUpdateFailed(error: Error) {
        guard case CardReaderServiceError.softwareUpdate(underlyingError: let underlyingError, batteryLevel: _) = error else {
            return
        }

        // Duplication of `readerSoftwareUpdateFailedBatteryLow` and `default is left to make factoring out easier later on.
        switch underlyingError {
        case .readerSoftwareUpdateFailedInterrupted:
            // Update was cancelled, don't treat this as an error
            return
        case .readerSoftwareUpdateFailedBatteryLow:
            alertsPresenter.present(
                viewModel: alertsProvider.updatingFailed(tryAgain: nil,
                                                         close: {
                    self.state = .searching
                }))
        default:
            alertsPresenter.present(
                viewModel: alertsProvider.updatingFailed(tryAgain: nil,
                                                         close: {
                    self.state = .searching
                }))
        }
    }

    private func showConnectionFailed(error: Error) {
        let retrySearch = {
            self.state = .retry
        }

        // TODO: Consider removing this in favour of retry only – continue doesn't make sense for a built-in reader
        let continueSearch = {
            self.state = .searching
        }

        let cancelSearch = {
            self.state = .cancel
        }

        guard case CardReaderServiceError.connection(let underlyingError) = error else {
            return alertsPresenter.present(
                viewModel: alertsProvider.connectingFailed(continueSearch: continueSearch, cancelSearch: cancelSearch))
        }

        switch underlyingError {
        case .incompleteStoreAddress(let adminUrl):
            alertsPresenter.present(
                viewModel: alertsProvider.connectingFailedIncompleteAddress(
                    openWCSettings: openWCSettingsAction(adminUrl: adminUrl,
                                                         retrySearch: retrySearch),
                    retrySearch: retrySearch,
                    cancelSearch: cancelSearch))
        case .invalidPostalCode:
            alertsPresenter.present(
                viewModel: alertsProvider.connectingFailedInvalidPostalCode(
                    retrySearch: retrySearch,
                    cancelSearch: cancelSearch))
        default:
            alertsPresenter.present(
                viewModel: alertsProvider.connectingFailed(
                    continueSearch: continueSearch,
                    cancelSearch: cancelSearch))
        }
    }

    private func openWCSettingsAction(adminUrl: URL?,
                                      retrySearch: @escaping () -> Void) -> ((UIViewController) -> Void)? {
        if let adminUrl = adminUrl {
            if let site = stores.sessionManager.defaultSite,
               site.isWordPressComStore {
                return { [weak self] viewController in
                    self?.openWCSettingsInWebview(url: adminUrl, from: viewController, retrySearch: retrySearch)
                }
            } else {
                return { [weak self] _ in
                    UIApplication.shared.open(adminUrl)
                    self?.showIncompleteAddressErrorWithRefreshButton()
                }
            }
        }
        return nil
    }
    private func openWCSettingsInWebview(url adminUrl: URL,
                                         from viewController: UIViewController,
                                         retrySearch: @escaping () -> Void) {
        let nav = NavigationView {
            AuthenticatedWebView(isPresented: .constant(true),
                                 url: adminUrl,
                                 urlToTriggerExit: nil,
                                 exitTrigger: nil)
                                 .navigationTitle(Localization.adminWebviewTitle)
                                 .navigationBarTitleDisplayMode(.inline)
                                 .toolbar {
                                     ToolbarItem(placement: .confirmationAction) {
                                         Button(action: {
                                             viewController.dismiss(animated: true) {
                                                 retrySearch()
                                             }
                                         }, label: {
                                             Text(Localization.doneButtonUpdateAddress)
                                         })
                                     }
                                 }
        }
        .wooNavigationBarStyle()
        let hostingController = UIHostingController(rootView: nav)
        viewController.present(hostingController, animated: true, completion: nil)
    }

    private func showIncompleteAddressErrorWithRefreshButton() {
        showConnectionFailed(error: CardReaderServiceError.connection(underlyingError: .incompleteStoreAddress(adminUrl: nil)))
    }

    /// An error occurred during discovery
    /// Presents the error in a modal
    ///
    private func onDiscoveryFailed(error: Error) {
        alertsPresenter.present(
            viewModel: alertsProvider.scanningFailed(error: error) { [weak self] in
            self?.returnFailure(error: error)
        })
    }

    /// Calls the completion with a success result
    ///
    private func returnSuccess(result: CardReaderConnectionResult) {
        onCompletion?(.success(result))
        state = .idle
    }

    /// Calls the completion with a failure result
    ///
    private func returnFailure(error: Error) {
        onCompletion?(.failure(error))
        state = .idle
    }
}

private extension BuiltInCardReaderConnectionController {
    enum Localization {
        static let adminWebviewTitle = NSLocalizedString(
            "WooCommerce Settings",
            comment: "Navigation title of the webview which used by the merchant to update their store address"
        )

        static let doneButtonUpdateAddress = NSLocalizedString(
            "Done",
            comment: "The button title to indicate that the user has finished updating their store's address and is" +
            "ready to close the webview. This also tries to connect to the reader again."
        )
    }
}
